{
TPictureClip
Use TPictureClip custom component to extract portions of a matrix of
pictures and assign the "cell picture" to another picture control.

Article:
. http://delphi.about.com/library/weekly/aa022900a.htm

********************************************
Zarko Gajic
About.com Guide to Delphi Programming
http://delphi.about.com
email: delphi.guide@about.com
********************************************
}

unit PictureClip;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, JPEG;

type
  TPictureClip = class(TImage)
  private
    FRows:Integer;
    FCols:Integer;
    FPicture:TPicture;
    function GetCellGraphic (Index:Integer):TPicture;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;override;
    property GraphicCell[Index:Integer]:TPicture read GetCellGraphic;
  published
    property Rows:Integer read FRows write FRows;
    property Cols:Integer read FCols write FCols;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TPictureClip]);
end;

constructor TPictureClip.Create(AOwner: TComponent);
begin
 inherited Create(AOwner);
 FPicture := TPicture.Create;
 Visible := False;
end;

destructor TPictureClip.Destroy;
begin
 FPicture.Destroy;
 inherited Destroy;
end;

function TPictureClip.GetCellGraphic (Index:Integer):TPicture;
var
 BWidth, BHeight : Integer;
 SrcRect, DestRect : TRect;
begin
 BWidth := Picture.Width div FCols;
 BHeight := Picture.Height div FRows;

 DestRect := Rect(0, 0, BWidth, BHeight);

 SrcRect.Left := (Index mod Cols) * BWidth;
 SrcRect.Top := (Index div Cols) * BHeight;
 SrcRect.Right := SrcRect.Left + BWidth;
 SrcRect.Bottom := SrcRect.Top + BHeight;

 FPicture.Bitmap.Width := BWidth;
 FPicture.Bitmap.Height := BHeight;
 FPicture.Bitmap.Canvas.CopyRect (DestRect, Canvas, SrcRect);

 GetCellGraphic := FPicture;
end;



end.
