unit _PuzzleTable;

interface

uses Windows, Classes, Types, ExtCtrls, Grids, Graphics, JPEG, Contnrs,
  PictureClip, Forms, Messages;

const
  cSplitCount = 4; // odnosno ukupan broj celja po x tj y kordinati

type
  TPuzzleObj = class( TObject)
  private
    FRB : integer;
    FTheImage: TImage;
  public
    constructor Create( iRB: integer); virtual;
    destructor Destroy; override;
    property RB: integer read FRB write FRB;
    property TheImage: TImage read FTheImage write FTheImage;
  end;

  TEmptyObj = class( TPuzzleObj)
  private
    FEmptyPoint: TPoint;
    procedure DrawImage;
  public
    constructor Create( iRB: integer); override;
    function IsAround(aPoint: TPoint): boolean;
    procedure SwapWithEmptyImage(oPuzzleObj: TPuzzleObj);
    property EmptyPoint: TPoint read FEmptyPoint write FEmptyPoint;
  end;

  TNumberPuzzle = class( TStringGrid)
  private
    FEmptyObj: TEmptyObj;
    FSplitCount: integer;
    FWithBorder: boolean;
    function GetRBCelije(X, Y: integer): integer;
    procedure NapraviObjekteCelija;
    procedure InicializujSlike( bWithBorder: boolean = true);
    procedure SetSplitCount(const Value: integer);
    procedure PromesajCelije;
    procedure JelSveSlozeno;
    procedure MoveCellImage( ACol, ARow: integer);
    procedure RunOnSelectCell(Sender: TObject; ACol, ARow: Longint;
       var CanSelect: Boolean);
    procedure CentrirajGrid;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
  protected
    procedure IscrtajSliku(iRB: integer; oImage: TImage); virtual;
    procedure InicializujGame; virtual;
    procedure DrawCell(ACol, ARow: Integer; ARect: TRect;
      AState: TGridDrawState); override;
  public
    constructor Create( AOwner: TComponent); override;
    destructor Destroy; override;
    property SplitCount: integer read FSplitCount write SetSplitCount;
    procedure Restart;
  end;

  TImagePuzzle = class (TNumberPuzzle)
  private
    FPictureClip: TPictureClip;
  protected
    procedure InicializujGame; override;
    procedure IscrtajSliku(iRB: integer; oImage: TImage); override;
  public
    constructor Create( AOwner: TComponent; sImageFileName: string); virtual;
    destructor Destroy; override;
  public
    procedure UcitajSliku( sFileName: string);
  end;

function MakePuzzle( bWithNumbers: boolean; sImageFileName: string): TNumberPuzzle;

implementation

uses
  SysUtils, Dialogs,
  StdCtrls // zbog ScrollBars := [ssNone];
  , Controls
  ;

{ TWorkArea }

const
  cCelsWidth = 70;

  cFrameBorder = 2; // za Frame3D

function MakePuzzle( bWithNumbers: boolean; sImageFileName: string): TNumberPuzzle;
begin
  if bWithNumbers then
    result := TNumberPuzzle.Create(nil)
  else
    result := TImagePuzzle.Create( nil, sImageFileName);
end;

constructor TPuzzleObj.Create( iRB: integer);
begin
  FRB := iRB;
  FTheImage := TImage.Create(nil);
end;

constructor TNumberPuzzle.Create( AOwner: TComponent);
begin
  inherited Create( AOwner);
  FixedCols := 0;
  FixedRows := 0;
  ScrollBars := ssNone;
  BorderStyle := bsNone;
  Ctl3D := false;

  DefaultDrawing := false;
  Options := Options - [goRangeSelect, goRangeSelect, goDrawFocusSelected,
    goVertLine, goHorzLine];
  OnSelectCell := RunOnSelectCell;

  FEmptyObj := TEmptyObj.Create(0);
  FSplitCount := cSplitCount;
  FWithBorder := true;
end;

destructor TNumberPuzzle.Destroy;
var
  i, j: integer;
begin
  if assigned( FEmptyObj ) then
    FEmptyObj.Free;

  for i:=0 to FSplitCount-1 do
    for j:=0 to FSplitCount -1 do
      if assigned( Objects[i, j]) then
        Objects[i, j].Free;
  inherited;
end;

procedure TNumberPuzzle.DrawCell(ACol, ARow: Integer; ARect: TRect;
  AState: TGridDrawState);
var
  oPO: TPuzzleObj;
begin
  if assigned( Objects[ACol, ARow]) then
  begin
    oPO := TPuzzleObj( Objects[ACol, ARow]);
    Canvas.StretchDraw( ARect, oPO.TheImage.Picture.Graphic );
  end;
end;

function TNumberPuzzle.GetRBCelije(X, Y: integer): integer;
begin
  Result := X * FSplitCount + Y
end;


procedure TNumberPuzzle.IscrtajSliku(iRB: integer; oImage: TImage);
var
  aRect: TRect;
begin
  with oImage  do
  begin
    aRect := ClientRect;
    if odd(iRB) then
      canvas.Brush.Color := clGray
    else
      canvas.Brush.Color := clSilver;
    canvas.FillRect( aRect);
    if FWithBorder then
      Frame3D( canvas, aRect, clWhite, clBlack, cFrameBorder);

    //canvas.Brush.Color := clGray;
    canvas.Pen.Color := clBlue;
    canvas.Font.Size := 50 ;
    canvas.Font.Style := [fsBold];
    canvas.Font.Name := 'Times new Roman';
    if iRB > 10 then
      canvas.TextOut( 17 + aRect.left , 8 + aRect.Top, inttostr( iRB+1))
    else
      canvas.TextOut( 30 + aRect.left, 8 + aRect.Top, inttostr( iRB+1));
  end;
end;

{ TPuzzleImage }

procedure TNumberPuzzle.Restart;
begin
  InicializujGame;
  CentrirajGrid;

  Perform( WM_SETREDRAW, 0, 0 ); // da se ne isrtava
  try
    PromesajCelije;
  finally
    Perform( WM_SETREDRAW, 1, 0 );
    invalidate
  end;
end;

procedure TNumberPuzzle.NapraviObjekteCelija;
var
  iRB: integer;
  iX, iY: integer;
  oPuzzleObj: TPuzzleObj;
begin
  for iX:=0 to FSplitCount-1 do
    for iY:=0 to FSplitCount-1 do
    begin
      iRB := GetRBCelije(iY, iX);
      oPuzzleObj := TPuzzleObj.Create( iRB);
      Objects[ iX, iY] := oPuzzleObj;
    end;
end;

constructor TEmptyObj.Create( iRB: integer);
begin
  inherited Create( iRB);
  FEmptyPoint := Point( -1, -1);
  DrawImage;
end;

function TEmptyObj.IsAround(aPoint: TPoint): boolean;
begin
  // aPontX je sigurno u granicama PuzzleTabele
  Result := false;
  if FEmptyPoint.Y = aPoint.Y then
    Result := ((FEmptyPoint.X - 1 = aPoint.X) or (FEmptyPoint.X + 1 = aPoint.X));
  if FEmptyPoint.X = aPoint.X then
    Result := ((FEmptyPoint.Y - 1 = aPoint.Y) or (FEmptyPoint.Y + 1 = aPoint.Y));
end;

procedure TEmptyObj.SwapWithEmptyImage(oPuzzleObj: TPuzzleObj);
var
  oImage: TImage;
begin
  self.RB := oPuzzleObj.RB;

  oImage := oPuzzleObj.TheImage;
  oPuzzleObj.TheImage := Self.TheImage;
  Self.TheImage := oImage;
end;

{ TEmptyImage }

procedure TEmptyObj.DrawImage;
var
  aRect: TRect;
begin
  if assigned( FTheImage) then
  begin
    with FTheImage do
    begin
      canvas.Brush.Color := clOlive;
      aRect := Rect(0, 0, Picture.Width, Picture.Height);
      canvas.FillRect( aRect);
      //arect := FTheImage.ClientRect;
      //Frame3D( canvas, aRect, clBlack, clWhite, cFrameBorder);
    end;
  end
end;

destructor TPuzzleObj.Destroy;
begin
  if assigned( FTheImage) then
    FreeAndNil( FTheImage);
  inherited;
end;

procedure TNumberPuzzle.InicializujSlike( bWithBorder: boolean = true);
var
  iX, iY: integer;
begin
  FWithBorder := bWithBorder;

  if Assigned( self.FEmptyObj) then
    FEmptyObj.DrawImage;

  for iX:=0 to FSplitCount-1 do
    for iY:=0 to FSplitCount-1 do
      if assigned( Objects[iY, iX]) then
        with TPuzzleObj( Objects[iY, iX]) do
        begin
          RB := self.GetRBCelije( iX, iY);
          IscrtajSliku( RB, TheImage )
        end;
end;

procedure TNumberPuzzle.InicializujGame;
var
  i, j: integer;
begin
  for i:=0 to FSplitCount -1 do
    for j:=0 to FSplitCount -1 do
      if assigned( Objects[i, j]) then
         TPuzzleObj(Objects[j, i]).Free;

  ColCount := FSplitCount;
  RowCount := FSplitCount;

  NapraviObjekteCelija;
  InicializujSlike;

  with FEmptyObj do
  begin
    EmptyPoint := Point( FSplitCount-1, FSplitCount-1);
    SwapWithEmptyImage(
       TPuzzleObj( Objects[EmptyPoint.X , EmptyPoint.Y]));
  end;
end;

procedure TNumberPuzzle.RunOnSelectCell(Sender: TObject; ACol, ARow: Longint;
   var CanSelect: Boolean);
var
  oPoint: TPoint;
begin
  oPoint := Point( ACol, ARow);
  if FEmptyObj.IsAround( oPoint) then
  begin
    MoveCellImage( ACol, ARow);
    JelSveSlozeno;
  end;
end;

procedure TNumberPuzzle.MoveCellImage( ACol, ARow: integer);
var
  oPuzzleObj: TPuzzleObj;
begin
  if Assigned( Objects[ACol, ARow]) then
  begin
    oPuzzleObj := TPuzzleObj(Objects[ACol, ARow]);
    with FEmptyObj.EmptyPoint do
    begin
      Objects[ACol, ARow] := Objects[X, Y];
      Objects[X, Y] := oPuzzleObj;
      X := ACol;
      Y := ARow;
    end;
  end;
end;

procedure TNumberPuzzle.SetSplitCount(const Value: integer);
begin
  FSplitCount := Value;
end;

procedure TNumberPuzzle.PromesajCelije;
// false = levo, true = desno
const
  cBrojPomeranja = 500;
const
  iKoef: array[boolean] of integer = (-1, 1);
var
  i: integer;
  oPoint: TPoint;
  iBrojMesta: array[boolean] of integer;
  bSmer : boolean;

  bPomeranjePoX: boolean;
  iStartMesto: integer; // trenutan polozaj X ili Y koridnate emtpy tacke
  iPomak: integer;
  M: integer;
begin
  for i:=0 to cBrojPomeranja do
  begin
                                              // PRIMERI VREDNOSTI:

    // prvo odredi da li ce da pomera po X ili po Y osi:
    bPomeranjePoX := boolean(random( 2));    // bPomeranjePoX := true

    // odredi polozaj EmptyPoint:
    with FEmptyObj do
    begin
      oPoint.X := EmptyPoint.X;
      oPoint.Y := EmptyPoint.Y;

      if bPomeranjePoX then
        iStartMesto := EmptyPoint.X           // iStartMesto := 3 (za X = npr 3)
      else
        iStartMesto := EmptyPoint.Y;

      bSmer :=  boolean( random(2));          // bSmer := false <--- ulevo
      iBrojMesta[false] := iStartMesto;       // iBrojMesta[false] := 3;
      iBrojMesta[true] :=                     // za FSplitCount = npr 5 =>
         (FSplitCount-1)-iStartMesto;         // iBrojMesta[true] := (5-1)-3 = 1;

      {iPomak :=                              //  iPomak := -1*random(3) =
        iKoef[bSmer] *                        //         := -1* 2 =
        random(iBrojMesta[bSmer]);            //         := -2      <--- 2 MESTA ULEVO!}

      iPomak := random(iBrojMesta[bSmer]);

      if iPomak = 0 then Continue;

      for M:=0 to iPomak do
      begin
        if bPomeranjePoX then
        begin
          oPoint.X := oPoint.X + iKoef[bSmer];
        end else begin
          oPoint.Y := oPoint.Y + iKoef[bSmer];
        end;
  
        MoveCellImage( oPoint.X, oPoint.Y );
      end;

    end;
  end;
end;

{ TImagePuzzle }

constructor TImagePuzzle.Create( AOwner: TComponent; sImageFileName: string);
begin
  inherited Create( AOwner);
  FPictureClip := TPictureClip.Create( nil);
  UcitajSliku( sImageFileName);
end;

destructor TImagePuzzle.Destroy;
begin
  if assigned( FPictureClip) then
    FPictureClip.Free;
  inherited;
end;

procedure TImagePuzzle.InicializujGame;
begin
  if assigned (FPictureClip) then
    with FPictureClip do
    begin
      Rows := FSplitCount;
      Cols := FSplitCount;
    end;

  inherited InicializujGame;
end;

procedure TImagePuzzle.IscrtajSliku(iRB: integer; oImage: TImage);
var
  aRect: TRect;
begin
  with oImage do
  begin
    Picture := FPictureClip.GraphicCell[iRB];
    aRect := Rect(0, 0, Picture.Width, Picture.Height);
    if self.FWithBorder then
      Frame3D( canvas, aRect, clWhite, clBlack, cFrameBorder);
  end;
end;

procedure TImagePuzzle.UcitajSliku(sFileName: string);
begin
  with FPictureClip do
    Picture.LoadFromFile(sFileName);
end;

procedure TNumberPuzzle.JelSveSlozeno;
var
  i, j: integer;
  iRB: integer;
  bReseno: boolean;
begin
  bReseno := true;

  for i:=0 to FSplitCount -1 do
    for j:=0 to FSplitCount -1 do
      if assigned( Objects[i, j]) then
      begin
        iRB := self.GetRBCelije( i, j);
        if (TPuzzleObj(Objects[j, i]).RB <> iRB) then
        begin
          bReseno := false;
          break;
        end;
      end;

  if bReseno then
  begin
    MessageDLG('CONGRATULATE!', mtInformation, [mbOk], 0);
    InicializujSlike( false);
    // vraca empty na svoje mesto:
    with FEmptyObj do
      EmptyPoint := Point(-1, -1);
    Invalidate;
  end;
end;

procedure TNumberPuzzle.CentrirajGrid;
var
  iSirina: integer;
begin
  DefaultColWidth := (Parent.ClientWidth - 8) div FSplitCount;
  DefaultRowHeight := DefaultColWidth;
  iSirina := DefaultColWidth * FSplitCount + FSplitCount;
  SetBounds( 3, 3, iSirina, iSirina);
end;

procedure TNumberPuzzle.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 0;
end;

end.
